import concurrent
import itertools
import operator
import sys
import threading
import typing
from concurrent.futures import ThreadPoolExecutor

from applitools.common import (
    DiffsFoundError,
    EyesError,
    NewTestError,
    TestFailedError,
    TestResultContainer,
    TestResults,
    TestResultsSummary,
    logger,
)
from applitools.common.utils import datetime_utils, iteritems
from applitools.core import EyesRunner

from .resource_cache import ResourceCache

if typing.TYPE_CHECKING:
    from typing import Optional, List, Dict
    from applitools.common import RenderingInfo
    from applitools.selenium.visual_grid import (
        RunningTest,
        VisualGridEyes,
        EyesConnector,
        VGTask,
    )


class VisualGridRunner(EyesRunner):
    def __init__(self, concurrent_sessions=None):
        # type: (Optional[int]) -> None
        super(VisualGridRunner, self).__init__()
        self._all_test_result = {}  # type: Dict[RunningTest, TestResults]

        kwargs = {}
        if sys.version_info >= (3, 6):
            kwargs["thread_name_prefix"] = "VGR-Executor"

        self.resource_cache = ResourceCache()  # type:ResourceCache
        self.put_cache = ResourceCache()  # type:ResourceCache
        self.all_eyes = []  # type: List[VisualGridEyes]
        self.still_running = True  # type: bool

        self._executor = ThreadPoolExecutor(max_workers=concurrent_sessions, **kwargs)
        self._rendering_info = None  # type: Optional[RenderingInfo]
        self._future_to_task = ResourceCache()  # type:ResourceCache
        thread = threading.Thread(target=self.run, args=())
        thread.setName(self.__class__.__name__)
        thread.daemon = True
        thread.start()
        self._thread = thread

    def __del__(self):
        self.stop()

    def aggregate_result(self, test, test_result):
        # type: (RunningTest, TestResults) -> None
        self._all_test_result[test] = test_result

    def render_info(self, eyes_connector):
        # type: (EyesConnector) -> RenderingInfo
        if self._rendering_info is None:
            self._rendering_info = eyes_connector.render_info()
        return self._rendering_info

    def open(self, eyes):
        # type: (VisualGridEyes) -> None
        self.all_eyes.append(eyes)
        self._rendering_info = eyes.rendering_info
        logger.debug("VisualGridRunner.open(%s)" % eyes)

    def run(self):
        logger.debug("VisualGridRunner.run()")
        while self.still_running:
            try:
                task = self.task_queue.pop()
                logger.debug("VisualGridRunner got task %s" % task)
            except IndexError:
                datetime_utils.sleep(1000)
                continue
            future = self._executor.submit(lambda task: task(), task)
            self._future_to_task[future] = task

    def stop(self):
        # type: () -> None
        logger.debug("VisualGridRunner.stop()")
        while sum(r.score for r in self.all_running_tests) > 0:
            datetime_utils.sleep(500)
        self.still_running = False
        for future in concurrent.futures.as_completed(self._future_to_task):
            task = self._future_to_task[future]
            try:
                future.result()
            except Exception as exc:
                logger.exception("%r generated an exception: %s" % (task, exc))
            else:
                logger.debug("%s task ran" % task)

        self.put_cache.executor.shutdown()
        self.resource_cache.executor.shutdown()
        self._executor.shutdown()
        self._thread.join()

    def get_all_test_results_impl(self, should_raise_exception=True):
        # type: (bool) -> TestResultsSummary
        while True:
            states = list(set([t.state for t in self.all_running_tests]))
            if len(states) == 1 and states[0] == "completed":
                break
            datetime_utils.sleep(500)

        all_results = []
        for test, test_result in iteritems(self._all_test_result):
            exception = None
            if test.test_result is None:
                exception = TestFailedError("Test haven't finished correctly")
            scenario_id_or_name = test_result.name
            app_id_or_name = test_result.app_name
            if test_result and test_result.is_unresolved and not test_result.is_new:
                exception = DiffsFoundError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
            if test_result and test_result.is_new:
                exception = NewTestError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
            if test_result and test_result.is_failed:
                exception = TestFailedError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
            all_results.append(
                TestResultContainer(test_result, test.browser_info, exception)
            )
            if exception and should_raise_exception:
                raise exception
        return TestResultsSummary(all_results)

    @property
    def all_running_tests(self):
        # type: ()-> List[RunningTest]
        return list(itertools.chain.from_iterable(e.test_list for e in self.all_eyes))

    @property
    def all_running_tests_by_score(self):
        # type: () -> List[RunningTest]
        return sorted(
            self.all_running_tests, key=operator.attrgetter("score"), reverse=True
        )

    @property
    def task_queue(self):
        # type: () -> List[VGTask]
        tests_to_run = self.all_running_tests_by_score
        if tests_to_run:
            test_to_run = tests_to_run[0]
            queue = test_to_run.queue
        else:
            queue = []
        return queue



from typing import TYPE_CHECKING

from applitools.common import TestResultContainer, TestResultsSummary
from applitools.core import EyesBase, EyesRunner

if TYPE_CHECKING:
    from typing import Optional, List
    from applitools.common import TestResults


class ClassicRunner(EyesRunner):
    def __init__(self):
        super(ClassicRunner, self).__init__()
        self._all_test_result = []  # type: List[TestResults]
        self.exception = None  # type: Optional[Exception]

    def get_all_test_results_impl(self, should_raise_exception=True):
        # type: (bool) -> TestResultsSummary
        if should_raise_exception and self.exception:
            raise self.exception
        result = []
        for test_result in self._all_test_result:
            result.append(TestResultContainer(test_result, None, None))
            EyesBase.log_session_results_and_raise_exception(
                should_raise_exception, test_result
            )
        return TestResultsSummary(result)

    def aggregate_result(self, test_result):
        # type: (TestResults) -> None
        self._all_test_result.append(test_result)



from applitools.common import (
    DeviceName,
    MatchLevel,
    Region,
    ScreenOrientation,
    StdoutLogger,
    logger,
    RectangleSize,
    TestResults,
    TestResultContainer,
    TestResultsSummary,
)
from applitools.common.config import BatchInfo  # noqa
from applitools.common.selenium import BrowserType, Configuration, StitchMode  # noqa
from applitools.common.server import FailureReports  # noqa
from applitools.core.cut import (  # noqa
    FixedCutProvider,
    NullCutProvider,
    UnscaledFixedCutProvider,
)

from .classic_runner import ClassicRunner  # noqa
from .eyes import Eyes  # noqa
from .fluent.target import Target  # noqa
from .visual_grid import VisualGridRunner  # noqa
from .webdriver import EyesWebDriver  # noqa
from .webelement import EyesWebElement  # noqa

__all__ = (
    # noqa
    "BatchInfo",
    "Region",
    "MatchLevel",
    "logger",
    "StdoutLogger",
    "Eyes",
    "Target",
    "FailureReports",
    "StitchMode",
    "VisualGridRunner",
    "BrowserType",
    "DeviceName",
    "Configuration",
    "ScreenOrientation",
    "FixedCutProvider",
    "NullCutProvider",
    "UnscaledFixedCutProvider",
    "ClassicRunner",
    "RectangleSize",
    "TestResults",
    "TestResultContainer",
    "TestResultsSummary",
)



from __future__ import absolute_import, unicode_literals

import json
import typing as tp
from collections import OrderedDict
from concurrent.futures.thread import ThreadPoolExecutor

import requests
import tinycss2

from applitools.common import logger
from applitools.common.utils import datetime_utils, general_utils
from applitools.common.utils.compat import urljoin
from applitools.selenium import eyes_selenium_utils
from applitools.selenium.positioning import ScrollPositionProvider

if tp.TYPE_CHECKING:
    from applitools.selenium.webdriver import EyesWebDriver

__all__ = ("get_full_window_dom",)
_CAPTURE_CSSOM_SCRIPT = """
function extractCssResources() {
    cssAndText = Array.from(document.querySelectorAll(
    'link[rel="stylesheet"],style')).map(el => {
        if (el.tagName.toUpperCase() === 'LINK') {
            return [null, el.getAttribute('href')];
        } else {
            return [el.textContent, null];
        }
    });
    return cssAndText
}
return extractCssResources();
"""
_CAPTURE_FRAME_SCRIPT = """
function captureFrame({ styleProps, attributeProps, rectProps, ignoredTagNames }) {
  const NODE_TYPES = {
    ELEMENT: 1,
    TEXT: 3,
  };

  function filter(x) {
    return !!x;
  }

  function notEmptyObj(obj) {
    return Object.keys(obj).length ? obj : undefined;
  }

  function iframeToJSON(el) {
    const obj = elementToJSON(el);
    try {
      if (el.contentDocument) {
        obj.childNodes = [captureNode(el.contentDocument.documentElement)];
      }
    } catch (ex) {
    } finally {
      return obj;
    }
  }

  function elementToJSON(el) {
    const tagName = el.tagName.toUpperCase();
    if (ignoredTagNames.indexOf(tagName) > -1) return null;
    const computedStyle = window.getComputedStyle(el);
    const boundingClientRect = el.getBoundingClientRect();

    const style = {};
    for (const p of styleProps) style[p] = computedStyle.getPropertyValue(p);

    const rect = {};
    for (const p of rectProps) rect[p] = boundingClientRect[p];

    const attributes = {};

    if (!attributeProps) {
      if (el.hasAttributes()) {
        var attrs = el.attributes;
        for (const p of attrs) {
          attributes[p.name] = p.value;
        }
      }
    }
    else {
      if (attributeProps.all) {
        for (const p of attributeProps.all) {
          if (el.hasAttribute(p)) attributes[p] = el.getAttribute(p);
        }
      }

      if (attributeProps[tagName]) {
        for (const p of attributeProps[tagName]) {
          if (el.hasAttribute(p)) attributes[p] = el.getAttribute(p);
        }
      }
    }
    return {
      tagName,
      style: notEmptyObj(style),
      rect: notEmptyObj(rect),
      attributes: notEmptyObj(attributes),
      childNodes: Array.prototype.map.call(el.childNodes, captureNode).filter(filter),
    };
  }

  function captureTextNode(node) {
    return {
      tagName: '#text',
      text: node.textContent,
    };
  }

  function captureNode(node) {
    switch (node.nodeType) {
      case NODE_TYPES.TEXT:
        return captureTextNode(node);
      case NODE_TYPES.ELEMENT:
        if (node.tagName.toUpperCase() === 'IFRAME') {
          return iframeToJSON(node);
        } else {
          return elementToJSON(node);
        }
      default:
        return null;
    }
  }

  return captureNode(document.documentElement);
}

return JSON.stringify(captureFrame(arguments[0]));
"""
_ARGS_OBJ = {
    "styleProps": [
        "background-color",
        "background-image",
        "background-size",
        "color",
        "border-width",
        "border-color",
        "border-style",
        "padding",
        "margin",
    ],
    "attributeProps": None,
    "rectProps": ["width", "height", "top", "left"],
    "ignoredTagNames": ["HEAD", "SCRIPT"],
}
CSS_DOWNLOAD_TIMEOUT = 3  # Secs


@datetime_utils.timeit
def get_full_window_dom(driver, return_as_dict=False):
    # type: (EyesWebDriver, bool) -> tp.Union[str, dict]

    dom_tree = json.loads(
        driver.execute_script(_CAPTURE_FRAME_SCRIPT, _ARGS_OBJ),
        object_pairs_hook=OrderedDict,
    )
    current_root_element = eyes_selenium_utils.curr_frame_scroll_root_element(driver)

    with eyes_selenium_utils.get_and_restore_state(
        ScrollPositionProvider(driver, current_root_element)
    ):
        logger.debug("Traverse DOM Tree")
        _traverse_dom_tree(driver, {"childNodes": [dom_tree], "tagName": "OUTER_HTML"})

    if return_as_dict:
        return dom_tree

    return json.dumps(dom_tree)


class DomNode(object):
    __slots__ = ("tag_name", "child_nodes", "is_html", "is_iframe")

    def __init__(self, tag_name, child_nodes):
        # type: (tp.Text, tp.List) -> None
        self.tag_name = tag_name
        self.child_nodes = child_nodes
        self.is_html = bool(self.tag_name == "HTML")
        self.is_iframe = bool(self.tag_name == "IFRAME")

    @classmethod
    def create_from_dom_tree(cls, node):
        # type: (tp.Dict) -> DomNode
        tag_name = node.get("tagName", "").upper()
        child_nodes = node.get("childNodes", [])
        return cls(tag_name, child_nodes)


def _traverse_dom_tree(driver, dom_tree):
    # type: (EyesWebDriver, tp.Dict) -> None
    """
    Walk through all IFRAMEs and add CSS to them
    """
    node = DomNode.create_from_dom_tree(dom_tree)
    if not node.tag_name:
        return None
    for index, sub_dom_tree in enumerate(_loop(driver, dom_tree)):
        # Reduce recursion optimization. Save from extra _loop calls
        if not sub_dom_tree["childNodes"]:
            continue
        with driver.switch_to.frame_and_back(index):
            _traverse_dom_tree(driver, sub_dom_tree)


def _loop(driver, dom_tree):
    # type: (EyesWebDriver, tp.Dict) -> tp.Iterable
    node = DomNode.create_from_dom_tree(dom_tree)
    if not node.child_nodes:
        return []

    def iterate_child_nodes(child_nodes):
        for sub_dom_tree in child_nodes:
            sub_node = DomNode.create_from_dom_tree(sub_dom_tree)
            if sub_node.is_iframe:
                yield sub_dom_tree
                continue

            if sub_node.is_html:
                sub_dom_tree["css"] = _get_frame_bundled_css(driver)
            if sub_node.child_nodes:
                # yield from iterate_child_nodes() in python 3
                for sub in iterate_child_nodes(sub_node.child_nodes):
                    yield sub

    return iterate_child_nodes(node.child_nodes)


@datetime_utils.timeit
def _get_frame_bundled_css(driver):
    # type: (EyesWebDriver) -> tp.Text
    base_url = driver.current_url  # type: ignore
    if not general_utils.is_absolute_url(base_url):
        logger.info("Base URL is not an absolute URL!")

    cssom_results = driver.execute_script(_CAPTURE_CSSOM_SCRIPT)
    raw_css_nodes = [
        CssNode.create(base_url, css_href, css_text)
        for css_text, css_href in cssom_results
    ]

    if len(raw_css_nodes) > 5:
        with ThreadPoolExecutor() as executor:
            results = executor.map(_process_raw_css_node, raw_css_nodes)
    else:
        results = [_process_raw_css_node(node) for node in raw_css_nodes]
    return "".join(results)


def _process_raw_css_node(node, minimize_css=True):
    # type: (CssNode, bool) -> tp.Text

    @datetime_utils.retry()
    def get_css(url):
        if url.startswith("blob:") or url.startswith("data:"):
            logger.warning("Passing blob URL: {}".format(url))
            return ""
        return requests.get(url, timeout=CSS_DOWNLOAD_TIMEOUT).text.strip()

    def iterate_css_sub_nodes(node, text=None):
        if text is None:
            text = node.text
            if node.text is None:
                text = get_css(node.url)

        for sub_node in _parse_and_serialize_css(node, text, minimize_css):
            if sub_node.url:
                text = get_css(sub_node.url)
                # yield from
                for res in iterate_css_sub_nodes(sub_node, text):
                    yield res
                continue
            yield sub_node.text

    return "".join(iterate_css_sub_nodes(node))


def _parse_and_serialize_css(node, text, minimize=False):
    # type: (CssNode, tp.Text, bool) -> tp.Generator
    def is_import_node(n):
        return n.type == "at-rule" and n.lower_at_keyword == "import"

    stylesheet = tinycss2.parse_stylesheet(
        text, skip_comments=True, skip_whitespace=True
    )
    for style_node in stylesheet:
        if is_import_node(style_node):
            for tag in style_node.prelude:
                if tag.type == "url":
                    logger.debug("The node has import")
                    yield CssNode.create_sub_node(parent_node=node, href=tag.value)
            continue

        try:
            if minimize:
                try:
                    # remove whitespaces inside blocks
                    style_node.content = [
                        tok for tok in style_node.content if tok.type != "whitespace"
                    ]
                except AttributeError as e:
                    logger.warning(
                        "Cannot serialize item: {}, cause error: {}".format(
                            style_node, str(e)
                        )
                    )
            serialized = style_node.serialize()
            if minimize:
                serialized = (
                    serialized.replace("\n", "").replace("/**/", " ").replace(" {", "{")
                )

        except TypeError as e:
            logger.warning(str(e))
            continue
        yield CssNode.create_serialized_node(text=serialized)


def _make_url(base_url, value):
    # type: (tp.Text, tp.Text) -> tp.Text
    if general_utils.is_absolute_url(
        value
    ) and not general_utils.is_url_with_scheme(  # noqa
        value
    ):
        url = urljoin("http://", value)
    else:
        url = urljoin(base_url, value)
    return url


class CssNode(object):
    __slots__ = ("base_url", "url", "text")

    def __init__(self, base_url, url, text):
        # type: (tp.Optional[tp.Text], tp.Optional[tp.Text], tp.Optional[tp.Text]) -> None
        self.base_url = base_url
        self.url = url
        self.text = text

    @classmethod
    def create(cls, base_url, href=None, text=None):
        # type: (tp.Text, tp.Optional[tp.Text], tp.Optional[tp.Text]) -> 'CssNode'
        url = _make_url(base_url, href) if href else None
        return cls(base_url, url, text)

    @classmethod
    def create_sub_node(cls, parent_node, href, text=None):
        # type: ('CssNode', tp.Text, tp.Optional[tp.Text]) -> 'CssNode'
        url = _make_url(parent_node.base_url, href)
        return cls(parent_node.base_url, url, text)

    @classmethod
    def create_serialized_node(cls, text):
        # type: (tp.Text) -> 'CssNode'
        return cls(base_url=None, url=None, text=text)



from enum import Enum

import attr
from ua_parser import user_agent_parser

__all__ = ("parse_user_agent_string", "BrowserNames", "OSNames", "UserAgent")


def parse_user_agent_string(ua_string):
    # type: (str) -> UserAgent
    parsed = user_agent_parser.Parse(ua_string)
    ua = parsed["user_agent"]
    os = parsed["os"]
    try:
        browser_name = BrowserNames(ua["family"])
    except ValueError:
        browser_name = BrowserNames.Unknown
    try:
        os_name = OSNames(os["family"])
    except ValueError:
        os_name = OSNames.Unknown
    return UserAgent(
        os=os_name,
        os_major_version=os["major"],
        os_minor_version=os["minor"],
        browser=browser_name,
        browser_major_version=ua["major"],
        browser_minor_version=ua["minor"],
    )


def to_float(val):
    try:
        return float(val)
    except TypeError:
        return -1


@attr.s
class UserAgent(object):
    os = attr.ib()
    os_major_version = attr.ib(converter=to_float)
    os_minor_version = attr.ib(converter=to_float)
    browser = attr.ib()
    browser_major_version = attr.ib(converter=to_float)
    browser_minor_version = attr.ib(converter=to_float)


class BrowserNames(Enum):
    Unknown = "Unknown"
    Edge = "Edge"
    IE = "IE"
    Firefox = "Firefox"
    Chrome = "Chrome"
    HeadlessChrome = "HeadlessChrome"
    Safari = "Safari"
    Chromium = "Chromium"
    MobileSafari = "Mobile Safari"


class OSNames(Enum):
    Unknown = "Unknown"
    Windows = "Windows"
    IOS = "iOS"
    Macintosh = "Macintosh"
    ChromeOS = "ChromeOS"
    Android = "Android"
    Linux = "Linux"



import codecs
import re
import sys
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))


def read(filename):
    return codecs.open(path.join(here, filename), "r", "utf-8").read()


# preventing ModuleNotFoundError caused by importing lib before installing deps
def get_version(package_name):
    version_file = read("applitools/{}/__version__.py".format(package_name))
    try:
        version = re.findall(r"^__version__ = \"([^']+)\"\r?$", version_file, re.M)[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")

    return version


install_requires = [
    "eyes-common=={}".format(get_version("selenium")),
    "eyes-core=={}".format(get_version("selenium")),
    "tinycss2==0.6.1",  # last version that supports python 2
    "selenium>=2.53.0,<4",
    "Pillow>=5.0,<7.0",  # 7 and above doesnt support python 2
    # uses for check if the mobile platform is used in the test
    "Appium-Python-Client>=0.4,<0.47",
    "ua_parser==0.8",
    "attrs>=19,<20",
    "transitions>=0.6,<0.7",
    "lxml>=4.4,<5",
]
# using this way of defining instead of 'typing>=3.5.2; python_version<="3.4"'
# for run on old version of setuptools without issues
if sys.version_info < (3, 5):
    # typing module was added as builtin in Python 3.5
    install_requires.append("typing >= 3.5.2")

if sys.version_info < (3,):
    install_requires.append("futures==3.2.0")

setup(
    name="eyes_selenium",
    version=get_version("selenium"),
    packages=find_packages(),
    url="http://www.applitools.com",
    author="Applitools Team",
    author_email="team@applitools.com",
    description="Applitools Python SDK. Selenium package",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
    ],
    keywords="applitools eyes eyes_selenium",
    install_requires=install_requires,
    package_data={
        "": ["README.rst", "LICENSE"],
        "applitools.selenium": ["py.typed", "resources/*.js"],
    },
    project_urls={
        "Bug Reports": "https://github.com/applitools/eyes.sdk.python/issues",
        "Source": "https://github.com/applitools/eyes.sdk.python/tree/master"
        "/eyes_selenium",
    },
)
